# -*- coding: utf-8 -*-
"""A very minimal Redis client library (not only) for MicroPython."""

try:
    import usocket as socket
except ImportError:
    import socket

try:
    import uselect as select
except ImportError:
    import select


CRLF = "\r\n"


class RedisError(Exception):
    """RESP error returned by the Redis server."""
    pass


class RedisTimeout(Exception):
    """Reply from the Redis server cannot be read within timeout."""


class ParseError(Exception):
    """Invalid input while parsing RESP data."""
    pass


def encode_request(*args):
    """Pack a series of arguments into a RESP array of bulk strings."""
    result = ["*" + str(len(args)) + CRLF]

    for arg in args:
        if arg is None:
            result.append('$-1' + CRLF)
        else:
            s = str(arg)
            result.append('$' + str(len(s)) + CRLF + s + CRLF)

    return "".join(result)


class Redis:
    """A very minimal Redis client."""

    def __init__(self, host='127.0.0.1', port=6379, timeout=3000, debug=False):
        self.debug = debug
        self._sock = None
        self._timeout = timeout
        self.connect(host, port)

    def connect(self, host=None, port=None):
        if host is not None:
            self._host = host

        if port is not None:
            self._port = port

        if not self._sock:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.connect(socket.getaddrinfo(self._host, self._port)[0][-1])
            self._sock_fd = self._sock.makefile('rb')
            try:
                self._sock_fd = self._sock_fd.fileno()
            except AttributeError:
                pass
            self._poller = select.poll()
            self._poller.register(self._sock_fd, select.POLLIN)

    def close(self):
        if self._sock:
            self._poller.unregister(self._sock_fd)
            self._sock.close()
            self._sock = None

    def do_cmd(self, cmd, *args):
        if not self._sock:
            raise RedisError("Not connected: use 'connect()' to connect to Redis server.")

        request = encode_request(cmd, *args)

        if self.debug:
            print("SEND: {!r}".format(request))

        self._sock.send(request.encode('utf-8'))
        return self._read_response()

    __call__ = do_cmd

    def __getattr__(self, name):
        if name.isalpha():
            return lambda *args: self.do_cmd(name, *args)

        raise AttributeError

    def _read_response(self):
        line = self._readuntil(lambda l, pos: l[-2:] == b'\r\n')
        rtype = line[:1].decode('utf-8')

        if rtype == '+':
            return line[1:-2]
        elif rtype == '-':
            raise RedisError(*line[1:-2].decode('utf-8').split(None, 1))
        elif rtype == ':':
            return int(line[1:-2])
        elif rtype == '$':
            length = int(line[1:-2])

            if length == -1:
                return None

            return self._readuntil(lambda l, pos: pos == length + 2)[:-2]
        elif rtype == '*':
            length = int(line[1:-2])

            if length == -1:
                return None

            return [self._read_response() for item in range(length)]
        else:
            raise ParseError("Invalid response header byte.")

    def _readuntil(self, predicate):
        buf = b''
        pos = 0
        while not predicate(buf, pos):
            readylist = self._poller.poll(self._timeout)
            if not readylist:
                raise RedisTimeout("Error reading response from Redis server within timeout.")

            for entry in readylist:
                if (entry[0] is self._sock_fd and entry[1] & select.POLLIN and not
                        entry[1] & (select.POLLHUP | select.POLLERR)):
                    buf += self._sock.recv(1)
                    pos += 1
                    break
            else:
                self.close()
                raise OSError("Error reading from socket.")

        if self.debug:
            print("RECV: {!r}".format(buf))

        return buf



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from io import open
from setuptools import setup


def read(*paths):
    with open(os.path.join(*paths), encoding='utf-8') as fp:
        return fp.read()


classifiers = """\
Development Status :: 3 - Alpha
Environment :: Other Environment
Intended Audience :: Developers
License :: OSI Approved :: MIT License
Operating System :: POSIX
Operating System :: POSIX :: Linux
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 2.7
Programming Language :: Python :: 3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
Programming Language :: Python :: 3.6
Programming Language :: Python :: Implementation :: CPython
Programming Language :: Python :: Implementation :: MicroPython
Programming Language :: Python :: Implementation :: PyPy
Topic :: Database
Topic :: Home Automation
Topic :: Software Development :: Libraries :: Python Modules
Topic :: System :: Networking
"""


setup(
    name='picoredis',
    version='0.1.0',
    author='Christopher Arndt',
    author_email='chris@chrisarndt.de',
    description='A very minimal Python Redis client library (not only) for MicroPython',
    long_description=read('README.rst'),
    license='MIT',
    url='https://github.com/SpotlightKid/micropython-stm-lib',
    keywords='database,micropython,redis,network',
    classifiers=[c for c in (c.strip() for c in classifiers.splitlines())
                 if c and not c.startswith('#')],
    py_modules=['picoredis'],
    package_dir={'': 'picoredis'},
    zip_safe=True
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest

from picoredis import protocol as p


CRLF = '\r\n'
GARBAGE = '§%$\r\n"%\n$/'
GLEN = len(GARBAGE)
MULTI_MESSAGE_STREAM = (
    '*3\r\n$3\r\nSET\r\n$15\r\nmemtier-8232902\r\n$2\r\nxx\r\n'
    '*3\r\n$3\r\nSET\r\n$15\r\nmemtier-8232902\r\n$2\r\nxx\r\n'
    '*3\r\n$3\r\nSET\r\n$15\r\nmemtier-7630684\r\n$3\r\nAAA\r\n'
)


def test_encode_request():
    assert p.encode_request("ping") == '*1\r\n$4\r\nping\r\n'
    assert p.encode_request("set", "foo", "bar") == (
        '*3\r\n$3\r\nset\r\n$3\r\nfoo\r\n$3\r\nbar\r\n')
    assert p.encode_request(42) == '*1\r\n$2\r\n42\r\n'
    assert p.encode_request(3.141) == '*1\r\n$5\r\n3.141\r\n'
    assert p.encode_request(None) == '*1\r\n$-1\r\n'


def test_decode():
    assert p.decode("+ping\r\n") == 'ping'
    assert p.decode(MULTI_MESSAGE_STREAM) == ["SET", "memtier-8232902", "xx"]


def test_decode_error():
    with pytest.raises(p.RedisError) as e:
        p.decode('-boom' + CRLF)

    assert e.value.args == ('boom',)

    with pytest.raises(p.RedisError) as e:
        p.decode('-boom shaka lakka' + CRLF)

    assert e.value.args == ('boom', 'shaka lakka')


def test_parse_stream():
    assert tuple(msg for msg, next in p.parse_stream(MULTI_MESSAGE_STREAM)) == (
        ["SET", "memtier-8232902", "xx"],
        ["SET", "memtier-8232902", "xx"],
        ["SET", "memtier-7630684", "AAA"]
    )


def test_parse_integer():
    assert p.parse_integer(':42' + CRLF)[0] == 42
    assert p.parse_integer(GARBAGE + ':42' + CRLF, GLEN)[0] == 42
    assert p.parse_integer(':42' + CRLF + GARBAGE)[0] == 42
    assert p.parse_integer(GARBAGE + ':42' + CRLF + GARBAGE, GLEN)[0] == 42

    with pytest.raises(p.ParseError) as e:
        p.parse_integer(':42')

    assert str(e.value) == "Unterminated integer after pos 1."

    with pytest.raises(p.ParseError) as e:
        p.parse_integer(':foo\r\n')

    assert str(e.value) == "Invalid integer at pos 1 - 4."


def test_parse_simple_string():
    assert p.parse_simple_string('+ping' + CRLF)[0] == 'ping'

    assert p.parse_simple_string(GARBAGE + '+ping' + CRLF, GLEN)[0] == 'ping'
    assert p.parse_simple_string('+ping' + CRLF + GARBAGE)[0] == 'ping'
    assert p.parse_simple_string(GARBAGE + '+ping' + CRLF + GARBAGE, GLEN)[0] == 'ping'

    with pytest.raises(p.ParseError) as e:
        p.parse_simple_string('+ping')

    assert str(e.value) == "Unterminated simple string after pos 1."


def test_parse_error():
    assert p.parse_error('-boom' + CRLF)[0] == ('boom',)
    assert p.parse_error(GARBAGE + '-boom' + CRLF, GLEN)[0] == ('boom',)
    assert p.parse_error('-boom' + CRLF + GARBAGE)[0] == ('boom',)
    assert p.parse_error(GARBAGE + '-boom' + CRLF + GARBAGE, GLEN)[0] == ('boom',)

    assert p.parse_error('-boom shaka lakka' + CRLF)[0] == ('boom', 'shaka lakka')
    assert p.parse_error(GARBAGE + '-boom shaka lakka' + CRLF, GLEN)[0] == ('boom', 'shaka lakka')
    assert p.parse_error('-boom shaka lakka' + CRLF + GARBAGE)[0] == ('boom', 'shaka lakka')
    assert p.parse_error(GARBAGE + '-boom shaka lakka' + CRLF + GARBAGE, GLEN)[0] == ('boom', 'shaka lakka')

    with pytest.raises(p.ParseError) as e:
        p.parse_error('-boom')

    assert str(e.value) == "Unterminated error string after pos 1."

    with pytest.raises(p.ParseError) as e:
        p.parse_error('-boom shaka lakka')

    assert str(e.value) == "Unterminated error string after pos 1."


def test_parse_bulk_string():
    assert p.parse_bulk_string('$12' + CRLF + 'hello world!' + CRLF)[0] == 'hello world!'
    assert p.parse_bulk_string('$14' + CRLF + 'hello\r\ngoodbye' + CRLF)[0] == 'hello\r\ngoodbye'
    assert p.parse_bulk_string('$-1' + CRLF)[0] is None
    assert p.parse_bulk_string(GARBAGE + '$-1' + CRLF, GLEN)[0] is None
    assert p.parse_bulk_string('$0' + CRLF + CRLF)[0] == ''
    assert p.parse_bulk_string(GARBAGE + '$12' + CRLF + 'hello world!' + CRLF, GLEN)[0] == 'hello world!'
    assert p.parse_bulk_string('$12' + CRLF + 'hello world!' + CRLF + GARBAGE)[0] == 'hello world!'
    assert p.parse_bulk_string(GARBAGE + '$12' + CRLF + 'hello world!' + CRLF + GARBAGE, GLEN)[0] == 'hello world!'

    with pytest.raises(p.ParseError) as e:
        p.parse_bulk_string('$14foo')

    assert str(e.value) == "Unterminated bulk string length after pos 1."

    with pytest.raises(p.ParseError) as e:
        p.parse_bulk_string('$foo\r\n')

    assert str(e.value) == "Invalid bulk string length at pos 1 - 4."

    with pytest.raises(p.ParseError) as e:
        p.parse_bulk_string('$1.5\r\n')

    assert str(e.value) == "Invalid bulk string length at pos 1 - 4."

    with pytest.raises(p.ParseError) as e:
        p.parse_bulk_string(GARBAGE + '$1.5\r\n', GLEN)

    assert str(e.value) == "Invalid bulk string length at pos 11 - 14."

    with pytest.raises(p.ParseError) as e:
        p.parse_bulk_string('$1.5\r\n' + GARBAGE)

    assert str(e.value) == "Invalid bulk string length at pos 1 - 4."

    with pytest.raises(p.ParseError) as e:
        p.parse_bulk_string('$14foo' + GARBAGE)

    assert str(e.value) == "Invalid bulk string length at pos 1 - 9."


def test_parse_array():
    assert p.parse_array('*3\r\n$3\r\nset\r\n$3\r\nfoo\r\n$3\r\nbar\r\n')[0] == ["set", "foo", "bar"]
    assert p.parse_array('*-1' + CRLF) is None


if __name__ == '__main__':
    pytest.main()
